
#include "Bitmap.h"

//basic constructor
Bitmap::Bitmap(){
    reset();
}

//constructor loads the bitmap when it is created
Bitmap::Bitmap(char *file){
    reset();
    loadBMP(file);
}

//destructor
Bitmap::~Bitmap(){
    if(data!=0) {
        delete[] data;
    }
}

const Bitmap& Bitmap::operator =(const Bitmap &other)
{
	loaded=other.loaded;
	width=other.width;
	height=other.height;
	bpp=other.bpp;
	if(data!=0) delete[] data;
	data=new unsigned char[3*height*width];
	for(int i=0; i<3*height*width; i++)
		data[i]=other.data[i];
	return *this;
}

//load a bitmap from a file and represent it correctly
//in memory
bool Bitmap::loadBMP(const char *file) {
    FILE *in;                  //file stream for reading
    char *tempData;       //temp storage for image data

    //bitmap is not loaded yet
    loaded=false;
    //make sure memory is not lost
    if(data!=0) {
        delete[] data;
    }

    //open the file for reading in binary mode
    fopen_s(&in, file,"rb");

    //if the file does not exist return in error
    if(in==0) {
        error="File not found";
        //fclose(in);
        return false;
    }

    //read in the entire BITMAPFILEHEADER
    fread(&bmfh,sizeof(BitmapFileHeader),1,in);
    //check for the magic number that says this is a bitmap
    if(bmfh.bfType!=BITMAP_MAGIC_NUMBER) {
        error="File is not in DIB format";
        fclose(in);
        return false;
    }

    //read in the entire BITMAPINFOHEADER
    fread(&bmih,sizeof(BitmapInfoHeader),1,in);

    //save the width, height and bits per pixel for external use
    width=bmih.biWidth;
    height=bmih.biHeight;
    bpp=bmih.biBitCount;

    //calculate the size of the image data with padding
    dataSize=(width*height*(unsigned int)(bmih.biBitCount/8.0));


    //if the bitmap is not 8 bits per pixel or more
    //return in error
    if(bpp<8) {
        error="File is not 24 bits per pixel";
        fclose(in);
        return false;
    }

    //set up the temporary buffer for the image data
    tempData=new char[dataSize];

    //exit if there is not enough memory
    if(tempData==0) {
        error="Not enough memory to allocate a temporary buffer";
        fclose(in);
        return false;
    }

    //read in the entire image
    fread(tempData,sizeof(char),dataSize,in);

    //close the file now that we have all the info
    fclose(in);

    //calculate the witdh of the final image in bytes
    byteWidth=padWidth=(width*bpp/8);

    //adjust the width for padding as necessary
    while(padWidth%4!=0) {
        padWidth++;
    }
    //change format from BGR to RGB
    if(bpp==24) {
    	loaded=convert(tempData);
   	}

    //clean up memory
    delete[] tempData;

    //return success
    return loaded;
}

//function to set the inital values
void Bitmap::reset(void) {
    loaded=false;
    data=0;
	height=0;
	width=0;
    error="";
}

bool Bitmap::convert(char* tempData) {
	int offset,diff;
	int delta=0;

	diff=width*height*3;
    //allocate the buffer for the final image data
    data=new unsigned char[diff];

    //exit if there is not enough memory
    if(data==0) {
        error="Not enough memory to allocate an image buffer";
        delete[] data;
        return false;
    }

    if(height>0) {
		offset=padWidth-byteWidth;
        //count backwards so you start at the front of the image
        for(unsigned int i=0;i<dataSize;i+=3) {
            //jump over the padding at the start of a new line
            if(i%byteWidth==0) {
				delta+=offset;
            }
            //transfer the data
            data[i + 2]=tempData[i+delta];
            data[i + 1]=tempData[i+1+delta];
            data[i]=tempData[i+2+delta];
        }
    }

    //image parser for a forward image
    else {
        offset=padWidth-byteWidth;
        int j=dataSize-3;
        for(unsigned int i=0;i<dataSize;i+=3) {
            //jump over the padding at the start of a new line
            if((i+1)%byteWidth==0) {
                i+=offset;
				delta+=offset;
            }
            //transfer the data
            *(data+j+2)=*(tempData+i+delta);
            *(data+j+1)=*(tempData+i+1+delta);
            *(data+j)=*(tempData+i+2+delta);
            j-=3;
        }
    }

    return true;
}

bool Bitmap::saveBMP(const char *filename)
{
	FILE* out;
	fopen_s(&out, filename,"wb");
	if(out==0)
		return false;
	BitmapFileHeader header;
	BitmapInfoHeader info;
	int bytewidth=width*3;
	int padwidth=bytewidth;
	while(padwidth%4!=0)
		padwidth++;
	//creating binary data for the image
	char* tempdata=(char*)malloc(height*padwidth);

    int offset=padwidth-bytewidth;
    //count backwards so you start at the front of the image
    for(int i=0;i<height*padwidth;i+=3) {
    //jump over the padding at the start of a new line
        if((i+1)%bytewidth==0) {
            for(int j=0; j<offset; j++)
				*(tempdata+i+j)=0;
			i+=offset;
        }
        //transfer the data
        *(tempdata+i+2)=*(data+i);
        *(tempdata+i+1)=*(data+i+1);
        *(tempdata+i)=*(data+i+2);
	}

	//prepare headers
	header.bfType=BITMAP_MAGIC_NUMBER;
	header.bfSize=sizeof(BitmapFileHeader)+sizeof(BitmapInfoHeader)+sizeof(tempdata);
	header.bfReserved1=header.bfReserved2=0;
	header.bfOffBits=sizeof(BitmapFileHeader)+sizeof(BitmapInfoHeader);

	info.biSize=40;
	info.biWidth=width;
	info.biHeight=height;
	info.biPlanes=1;
	info.biBitCount=24;
	info.biCompression=0;
	info.biSizeImage=0;
	info.biXPelsPerMeter=4000;
	info.biYPelsPerMeter=4000;
	info.biClrUsed=0;
	info.biClrImportant=0;
	
	//write data to file
	
	fwrite(&header,sizeof(BitmapFileHeader),1,out);
	fwrite(&info,sizeof(BitmapInfoHeader),1,out);
	fwrite(tempdata,height*padwidth,1,out);

	fclose(out);
	free(tempdata);
	return true;

}