#include "CConvolution3x3.h"
#include "Common.h"
#include "CTimer.h"

#include "Pfm.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CConvolution3x3

CConvolution3x3::CConvolution3x3(const char* FileName, size_t TileSize[2], float ConvKernel[3][3],
	bool Monochrome, float Offset)
	:m_FileName(FileName),
	m_Height(0), m_Width(0),
	m_Monochrome(Monochrome),
	m_Offset(Offset),
	m_dKernelConstants(NULL),
	m_Program(NULL), m_ConvolutionKernel(NULL)
{
	m_TileSize[0] = TileSize[0];
	m_TileSize[1] = TileSize[1];

	m_KernelWeight = 0;
	for(int i = 0; i < 3; i++)
	{
		m_hSourceChannels[i] = NULL;
		m_hCPUResultChannels[i] = NULL;
		m_hGPUResultChannels[i] = NULL;
		
		m_dSourceChannels[i] = NULL;
		m_dResultChannels[i] = NULL;

		for(int j = 0; j < 3; j++)
		{
			m_hConvolutionKernel[i][j] = ConvKernel[i][j];
			m_KernelWeight += ConvKernel[i][j];
		}
	}

	if(m_KernelWeight > 0)
		m_KernelWeight = 1.0f / m_KernelWeight;
	else
		m_KernelWeight = 1.0f;
}

CConvolution3x3::~CConvolution3x3()
{
}

bool CConvolution3x3::InitResources(cl_device_id Device, cl_context Context)
{
	PFM inputPfm;
	if (!inputPfm.LoadRGB(m_FileName.c_str())) {
		cout<<"Error loading file: " << m_FileName.c_str() << "." << endl;
		return false;
	}

	//internally, we convert the bitmap to floats, and execute the same convolution
	//operation on its three channels separately
	m_Height = inputPfm.height;
	m_Width = inputPfm.width;
	m_Pitch = m_Width;
	if(m_Width % 32 != 0)
		m_Pitch = m_Width + 32 - (m_Width % 32); //This will make sure that the data accesses are ALWAYS coalesced

	cout<<"Size of image: "<<m_Width<<" x "<<m_Height<<endl;

	//allocate data for the float channels
	for(int i = 0; i < 3; i++)
	{
		m_hSourceChannels[i] = new float[m_Height * m_Pitch];
		m_hCPUResultChannels[i] = new float[m_Height * m_Pitch];
		m_hGPUResultChannels[i] = new float[m_Height * m_Pitch];
	}

	//extract R, G, B channels
	unsigned int pixelOffset = 0;
	unsigned int trippleOffset = 0;
	for(unsigned int y = 0; y < m_Height; y++)
	{
		for(unsigned int x = 0; x < m_Width; x++)
		{
			m_hSourceChannels[0][pixelOffset] = inputPfm.pImg[trippleOffset    ];
			m_hSourceChannels[1][pixelOffset] = inputPfm.pImg[trippleOffset + 1];
			m_hSourceChannels[2][pixelOffset] = inputPfm.pImg[trippleOffset + 2];

			//monochrome: the data is converted to grayscale
			if(m_Monochrome)
				RGBToGrayScale(	m_hSourceChannels[0][pixelOffset],
								m_hSourceChannels[1][pixelOffset],
								m_hSourceChannels[2][pixelOffset],
								m_hSourceChannels[0][pixelOffset]);

			pixelOffset++;
			trippleOffset += 3;
		}
		//pad the image with zeros
		for(unsigned int i = 0; i < m_Pitch - m_Width; i++)
		{
			m_hSourceChannels[0][pixelOffset + i] = 0.0f;
			m_hSourceChannels[1][pixelOffset + i] = 0.0f;
			m_hSourceChannels[2][pixelOffset + i] = 0.0f;
		}
		pixelOffset += m_Pitch - m_Width;
	}
	//create GPU resources
	
	//we can init the kernel buffer during creation as its contents will not change
	cl_int clError;
	cl_float kernelConstants[11];
	for(int y = 0; y < 3; y++)
		for(int x = 0; x < 3; x++)
			kernelConstants[3*y + x] = m_hConvolutionKernel[y][x];
	kernelConstants[9] = m_KernelWeight;
	kernelConstants[10] = m_Offset;

	m_dKernelConstants = clCreateBuffer(Context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, 11 * sizeof(cl_float), 
		kernelConstants, &clError);
	V_RETURN_FALSE_CL(clError, "Error allocating device kernel constants.");

	unsigned int dataSize = m_Pitch * m_Height * sizeof(cl_float);
	
	for(int i = 0; i < 3; i++)
	{
		m_dSourceChannels[i] = clCreateBuffer(Context, CL_MEM_READ_ONLY, dataSize, NULL, &clError);
		V_RETURN_FALSE_CL(clError, "Error allocating device input array");

		m_dResultChannels[i] = clCreateBuffer(Context, CL_MEM_WRITE_ONLY, dataSize, NULL, &clError);
		V_RETURN_FALSE_CL(clError, "Error allocating device output array");
	}

	char* programCode = NULL;
	size_t programSize = 0;

	LoadProgram("Convolution3x3.cl", &programCode, &programSize);

	//create a program object
	m_Program = clCreateProgramWithSource(Context, 1, (const char**) &programCode, &programSize, &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create program from file.");

	//build program
	clError = clBuildProgram(m_Program, 1, &Device, NULL, NULL, NULL);
	if(clError != CL_SUCCESS)
	{
		PrintBuildLog(m_Program, Device);
		return false;
	}

	//create kernel(s)
	m_ConvolutionKernel = clCreateKernel(m_Program, "Convolution", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create kernel.");
	
	//bind kernel attributes
	clError = clSetKernelArg(m_ConvolutionKernel, 2, sizeof(cl_mem), (void*)&m_dKernelConstants);
	clError |= clSetKernelArg(m_ConvolutionKernel, 3, sizeof(cl_uint), (void*)&m_Width);
	clError |= clSetKernelArg(m_ConvolutionKernel, 4, sizeof(cl_uint), (void*)&m_Height);
	clError |= clSetKernelArg(m_ConvolutionKernel, 5, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting kernel arguments");

	return true;
}

void CConvolution3x3::ReleaseResources()
{
	// host data
	for(int i = 0; i < 3; i++)
	{
		SAFE_DELETE_ARRAY( m_hSourceChannels[i] );
		SAFE_DELETE_ARRAY( m_hCPUResultChannels[i] );
		SAFE_DELETE_ARRAY( m_hGPUResultChannels[i] );

		SAFE_RELEASE_MEMOBJECT(m_dSourceChannels[i]);
		SAFE_RELEASE_MEMOBJECT(m_dResultChannels[i]);
	}

	SAFE_RELEASE_MEMOBJECT(m_dKernelConstants);

	SAFE_RELEASE_KERNEL(m_ConvolutionKernel);
	SAFE_RELEASE_PROGRAM(m_Program);
}

void CConvolution3x3::ComputeGPU(cl_context Context, cl_command_queue CommandQueue)
{
	CTimer timer;
	int nIterations = 10;

	//do 1 or 3 convolution steps, based on the number of color channels to process
	unsigned int numChannels = m_Monochrome ? 1 : 3;

	size_t dataSize = m_Pitch * m_Height * sizeof(cl_float);

	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)
	{
		//copy input data to the GPU
		V_RETURN_CL( clEnqueueWriteBuffer(CommandQueue, m_dSourceChannels[iChannel], CL_FALSE, 0, dataSize, 
												m_hSourceChannels[iChannel], 0, NULL, NULL), "Error copying input data to device!" );
	}
	
	clFinish(CommandQueue);
	timer.Start();

	//perform the convolution and measure the performance
	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)	
		ConvolutionChannelGPU(iChannel, Context, CommandQueue, nIterations);

	clFinish(CommandQueue);
	timer.Stop();

	double ms = 1000 * timer.GetElapsedTime() / double(nIterations);
	cout<<"  Average time: "<<ms<<" ms, throughput: "<< 1.0e-6 * (double)numChannels * m_Width * m_Height / ms << " Gpixels/s" <<endl;

	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)
	{
		//copy the results back to the CPU
		V_RETURN_CL( clEnqueueReadBuffer(CommandQueue, m_dResultChannels[iChannel], CL_TRUE, 0, dataSize,
									m_hGPUResultChannels[iChannel], 0, NULL, NULL), "Error reading back results from the device!" );
	}


	SaveImage("GPUResult3x3.pfm", m_hGPUResultChannels);

}

void CConvolution3x3::ConvolutionChannelGPU(unsigned int Channel, cl_context Context, cl_command_queue CommandQueue, int NIterations)
{
	
	//also measure the time for the first channel
	CTimer timer;
	
	clFinish(CommandQueue);
	timer.Start();

	size_t globalWorkSize[2] = {GetGlobalWorkSize(m_Width, m_TileSize[0]), GetGlobalWorkSize(m_Height, m_TileSize[1])};
	
	cl_int clErr;
	clErr  = clSetKernelArg(m_ConvolutionKernel, 0, sizeof(cl_mem), (void*)&m_dResultChannels[Channel]);
	clErr |= clSetKernelArg(m_ConvolutionKernel, 1, sizeof(cl_mem), (void*)&m_dSourceChannels[Channel]);
	V_RETURN_CL(clErr, "Error setting kernel arguments!");

	for(int i = 0; i < NIterations; i++)
	{
		clErr = clEnqueueNDRangeKernel(CommandQueue, m_ConvolutionKernel, 2, NULL, globalWorkSize, m_TileSize, 0, NULL, NULL);
		V_RETURN_CL(clErr, "Error executing kernel!");
	}

}

void CConvolution3x3::ConvolutionChannelCPU(unsigned int Channel)
{
	//also measure the time for the first channel
	CTimer timer;
	
	timer.Start();
	
		for(unsigned int y = 0; y < m_Height; y++)
		{
			for(unsigned int x = 0; x < m_Width; x++)
			{
				float value = 0;
				//apply convolution kernel
				for(int offsetY = -1; offsetY < 2; offsetY ++)
				{
					int sy = y + offsetY;
					if(sy >= 0 && sy < int(m_Height))
						for(int offsetX = -1; offsetX < 2; offsetX++)
						{
							int sx = x + offsetX;
							if(sx >= 0 && sx < int(m_Width))
								value += m_hSourceChannels[Channel][sy * m_Pitch + sx] * m_hConvolutionKernel[1 + offsetY][1 + offsetX];
						}
				}
				m_hCPUResultChannels[Channel][y * m_Pitch + x] = value * m_KernelWeight + m_Offset;		
			}
		}

	timer.Stop();

	if(Channel == 0)
	{
		double ms = 1000 * timer.GetElapsedTime();
		cout<<"  Average time: "<<ms<<" ms, throughput: "<< 1.0e-6 * (double)m_Width * m_Height / ms << " Gpixels/s" <<endl;
	}
}

void CConvolution3x3::ComputeCPU()
{
	//number of channels to compute
	unsigned int numChannels = m_Monochrome ? 1 : 3;

	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)
	{
		ConvolutionChannelCPU(iChannel);
	}

	SaveImage("CPUResult3x3.pfm", m_hCPUResultChannels);
}

void CConvolution3x3::SaveImage(const char* FileName, float* Channels[3])
{
	//save the result back to the disk
	// Write data to the disc
	PFM resPfm;
	resPfm.pImg = new float[m_Width * m_Height * 3];
	unsigned int pfmOffset = 0;
	unsigned int pixOffset = 0;
	for(unsigned int y = 0; y < m_Height; y++)
	{
		for(unsigned int x = 0; x < m_Width; x++)
		{
			if(m_Monochrome)
			{
				resPfm.pImg[pfmOffset] = Channels[0][pixOffset];
				resPfm.pImg[pfmOffset + 1] = Channels[0][pixOffset];
				resPfm.pImg[pfmOffset + 2] = Channels[0][pixOffset];
			}
			else
			{
				resPfm.pImg[pfmOffset] = Channels[0][pixOffset];
				resPfm.pImg[pfmOffset + 1] = Channels[1][pixOffset];
				resPfm.pImg[pfmOffset + 2] = Channels[2][pixOffset];
			}
			pfmOffset += 3;
			pixOffset++;
		}
		pixOffset += m_Pitch - m_Width;
	}
	resPfm.width = m_Width;
	resPfm.height = m_Height;
	if(!resPfm.SaveRGB(FileName))
	{
		cout<<"Error saving "<<FileName<<"."<<endl;
		return;
	}
}

bool CConvolution3x3::ValidateResults()
{
	//number of channels to compute
	unsigned int numChannels = m_Monochrome ? 1 : 3;

	//calculate the average squared difference
	float avgError = 0;
	float maxError = 0;
	float scaling = 1.0f / float(numChannels * m_Width * m_Height);

	for(unsigned int y = 0; y < m_Height; y++)
		for(unsigned int x = 0; x < m_Width; x++)
			for(unsigned int i = 0; i < numChannels; i++)
			{
				float L2Error = m_hCPUResultChannels[i][y * m_Pitch + x] - m_hGPUResultChannels[i][y * m_Pitch + x];
				L2Error = L2Error * L2Error;
				maxError = max(maxError, L2Error);
				avgError += L2Error * scaling;

				//temp: to see the difference
				m_hCPUResultChannels[i][y * m_Pitch + x] = L2Error;
			}
	cout<<"Average sq. difference: "<<avgError<<endl;
	cout<<"Maximum sq. difference: "<<maxError<<endl;

	//save difference image
	SaveImage("DifferenceImage3x3.pfm", m_hCPUResultChannels);

	return avgError < 0.0001f;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
