#include "CConvolutionBilateral.h"
#include "Bitmap.h"
#include "CTimer.h"


//----------------------------------------------------------------------------------------
#define DEPTH_THRESHOLD	0.025f
#define NORM_THRESHOLD	0.9f

// These two functions define 'normal' and 'depth' discontinuities
bool IsNormalDiscontinuity(const cl_float4 &n1, const cl_float4 &n2){
	return abs(n1.s[0] * n2.s[0] + n1.s[1] * n2.s[1] + n1.s[2] * n2.s[2]) < NORM_THRESHOLD;
}

bool IsDepthDiscontinuity(float d1, float d2){
	return fabs(d1 - d2) > DEPTH_THRESHOLD;
}

//-------------------------------------------------------------------------------------------------------------------

// CConvolutionSeparable
CConvolutionBilateral::CConvolutionBilateral(const char* FileName, const char* NormalFileName, const char* DepthFileName, size_t LocalSizeHorizontal[2], size_t LocalSizeVertical[2],
		int StepsHorizontal, int StepsVertical, int KernelRadius, float* pKernelHorizontal, float* pKernelVertical)
		:m_FileName(FileName), m_NormalFileName(NormalFileName), m_DepthFileName(DepthFileName),
		m_Height(0), m_Width(0), m_KernelRadius(KernelRadius), m_StepsHorizontal(StepsHorizontal), m_StepsVertical(StepsVertical),
		m_hCPUDiscBuffer(NULL), m_hGPUDiscBuffer(NULL), m_hNormDepthBuffer(NULL),
		m_dDiscBuffer(NULL), m_dNormDepthBuffer(NULL),
		m_HorizontalDiscKernel(NULL), m_VerticalDiscKernel(NULL),
		m_dKernelHorizontal(NULL), m_dKernelVertical(NULL),
		m_Program(NULL), m_HorizontalKernel(NULL), m_VerticalKernel(NULL)
{
	m_LocalSizeHorizontal[0] = LocalSizeHorizontal[0];
	m_LocalSizeHorizontal[1] = LocalSizeHorizontal[1];
	m_LocalSizeVertical[0] = LocalSizeVertical[0];
	m_LocalSizeVertical[1] = LocalSizeVertical[1];

	const unsigned int kernelSize = 2 * m_KernelRadius + 1;
	m_hKernelHorizontal = new float[kernelSize];
	m_hKernelVertical = new float[kernelSize];
	memcpy(m_hKernelHorizontal, pKernelHorizontal, kernelSize * sizeof(float));
	memcpy(m_hKernelVertical, pKernelVertical, kernelSize * sizeof(float));

	for(int i = 0; i < 3; i++)
	{
		m_hSourceChannels[i] = NULL;
		m_hCPUResultChannels[i] = NULL;
		m_hGPUResultChannels[i] = NULL;

		m_dResultChannel[i] = NULL;
		m_dResultChannelTemp[i] = NULL;
	}
}

CConvolutionBilateral::~CConvolutionBilateral()
{
	delete [] m_hKernelHorizontal;
	delete [] m_hKernelVertical;
}

bool CConvolutionBilateral::InitResources(cl_device_id Device, cl_context Context)
{
	// Load PFM files
	PFM inputPfm;
	if (!inputPfm.LoadRGB(m_FileName.c_str())) {
		cout<<"Error loading file: " << m_FileName.c_str() << "." << endl;
		return false;
	}

	PFM normalsPfm;
	if (!normalsPfm.LoadRGB(m_NormalFileName.c_str())) {
		cout<<"Error loading file: " << m_NormalFileName.c_str() << "." << endl;
		return false;
	}

	PFM depthPfm;
	if (!depthPfm.LoadRGB(m_DepthFileName.c_str())) {
		cout<<"Error loading file: " << m_DepthFileName.c_str() << "." << endl;
		return false;
	}

	//internally, we convert the bitmap to floats, and execute the same convolution
	//operation on its three channels separately
	m_Height = inputPfm.height;
	m_Width = inputPfm.width;
	m_Pitch = m_Width;
	if(m_Width % 32 != 0)
		m_Pitch = m_Width + 32 - (m_Width % 32); //This will make sure that the data accesses are ALWAYS coalesced

	cout<<"Size of the input: "<<m_Width<<" x "<<m_Height<<endl;

	//allocate data for the float channels
	for(int i = 0; i < 3; i++) {
		m_hSourceChannels[i]	= new float[m_Height * m_Pitch];
		m_hCPUResultChannels[i] = new float[m_Height * m_Pitch];
		m_hGPUResultChannels[i] = new float[m_Height * m_Pitch];
	}


	m_hCPUDiscBuffer	= new cl_int[m_Height * m_Pitch];
	m_hGPUDiscBuffer	= new cl_int[m_Height * m_Pitch];

	m_hNormDepthBuffer	= new cl_float4[m_Height * m_Pitch];

	unsigned int pixelOffset = 0;
	unsigned int trippleOffset = 0;
	for(int y = 0; y < m_Height; y++) {
		for(int x = 0; x < m_Width; x++) {
			m_hNormDepthBuffer[pixelOffset].s[0] = normalsPfm.pImg[trippleOffset    ];
			m_hNormDepthBuffer[pixelOffset].s[1] = normalsPfm.pImg[trippleOffset + 1];
			m_hNormDepthBuffer[pixelOffset].s[2] = normalsPfm.pImg[trippleOffset + 2];
			m_hNormDepthBuffer[pixelOffset].s[3] = depthPfm.pImg[trippleOffset];

			m_hSourceChannels[0][pixelOffset] = inputPfm.pImg[trippleOffset    ];
			m_hSourceChannels[1][pixelOffset] = inputPfm.pImg[trippleOffset + 1];
			m_hSourceChannels[2][pixelOffset] = inputPfm.pImg[trippleOffset + 2];

			pixelOffset++;
			trippleOffset += 3;
		}
		pixelOffset += m_Pitch - m_Width;
	}


	//create GPU resources
	//we can init the kernel buffer during creation as its contents will not change
	const unsigned int kernelSize = 2 * m_KernelRadius + 1;

	cl_int clError = 0;
	cl_int clErr;
	m_dKernelHorizontal = clCreateBuffer(Context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, kernelSize * sizeof(cl_float),  m_hKernelHorizontal, &clErr);
	clError |= clErr;
	m_dKernelVertical = clCreateBuffer(Context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, kernelSize * sizeof(cl_float),  m_hKernelVertical, &clErr);
	clError |= clErr;
	V_RETURN_FALSE_CL(clError, "Error allocating device kernel constants.");

	m_dDiscBuffer = clCreateBuffer(Context, CL_MEM_READ_WRITE, m_Pitch * m_Height * sizeof(cl_int),  NULL, &clErr);
	clError = clErr;
	m_dNormDepthBuffer = clCreateBuffer(Context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, m_Pitch * m_Height * sizeof(cl_float4),  m_hNormDepthBuffer, &clErr);
	clError = clErr;
	V_RETURN_FALSE_CL(clError, "Error allocating device memory.");


	for(int i = 0; i < 3; i++) {
		m_dResultChannelTemp[i] = clCreateBuffer(Context, CL_MEM_READ_WRITE, m_Pitch * m_Height * sizeof(cl_float), NULL, &clError);
		V_RETURN_FALSE_CL(clError, "Error allocating device input array");
		m_dResultChannel[i] = clCreateBuffer(Context, CL_MEM_READ_WRITE, m_Pitch * m_Height * sizeof(cl_float), NULL, &clError);
		V_RETURN_FALSE_CL(clError, "Error allocating device input array");
	}

	char* programCode = NULL;
	size_t programSize = 0;

	LoadProgram("ConvolutionBilateral.cl", &programCode, &programSize);

	//create a program object
	m_Program = clCreateProgramWithSource(Context, 1, (const char**) &programCode, &programSize, &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create program from file.");

	//build program
	
	//This time we define several kernel-specific constants that we did not know during
	//implementing the kernel, but we need to include during compile time.
	char compileOptions[2048];
	sprintf_s(compileOptions, "-cl-fast-relaxed-math " \
		"-D KERNEL_RADIUS=%d " \
		"-D H_GROUPSIZE_X=%d -D H_GROUPSIZE_Y=%d -D H_RESULT_STEPS=%d " \
		"-D V_GROUPSIZE_X=%d -D V_GROUPSIZE_Y=%d -D V_RESULT_STEPS=%d ",
		m_KernelRadius, 
		m_LocalSizeHorizontal[0], m_LocalSizeHorizontal[1], m_StepsHorizontal,
		m_LocalSizeVertical[0], m_LocalSizeVertical[1], m_StepsVertical);
	
	clError = clBuildProgram(m_Program, 1, &Device, compileOptions, NULL, NULL);
	if(clError != CL_SUCCESS)
	{
		PrintBuildLog(m_Program, Device);
		return false;
	}

	//create kernel(s)

	m_HorizontalDiscKernel = clCreateKernel(m_Program, "DiscontinuityHorizontal", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create horizontal discontinuity detection kernel.");

	m_VerticalDiscKernel = clCreateKernel(m_Program, "DiscontinuityVertical", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create vertical discontinuity detection kernel.");
	
	m_HorizontalKernel = clCreateKernel(m_Program, "ConvHorizontal", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create horizontal kernel.");
	
	m_VerticalKernel = clCreateKernel(m_Program, "ConvVertical", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create vertical kernel.");

	//bind kernel attributes
	clError  = clSetKernelArg(m_HorizontalDiscKernel, 0, sizeof(cl_mem), (void*)&m_dDiscBuffer);
	clError |= clSetKernelArg(m_HorizontalDiscKernel, 1, sizeof(cl_mem), (void*)&m_dNormDepthBuffer);
	clError |= clSetKernelArg(m_HorizontalDiscKernel, 2, sizeof(cl_uint), (void*)&m_Width);
	clError |= clSetKernelArg(m_HorizontalDiscKernel, 3, sizeof(cl_uint), (void*)&m_Height);
	clError |= clSetKernelArg(m_HorizontalDiscKernel, 4, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting horizontal discontinuity kernel arguments");

	clError  = clSetKernelArg(m_VerticalDiscKernel, 0, sizeof(cl_mem), (void*)&m_dDiscBuffer);
	clError |= clSetKernelArg(m_VerticalDiscKernel, 1, sizeof(cl_mem), (void*)&m_dNormDepthBuffer);
	clError |= clSetKernelArg(m_VerticalDiscKernel, 2, sizeof(cl_uint), (void*)&m_Width);
	clError |= clSetKernelArg(m_VerticalDiscKernel, 3, sizeof(cl_uint), (void*)&m_Height);
	clError |= clSetKernelArg(m_VerticalDiscKernel, 4, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting vertical discontinuity kernel arguments");

	clError  = clSetKernelArg(m_HorizontalKernel, 2, sizeof(cl_mem), (void*)&m_dDiscBuffer);
	clError |= clSetKernelArg(m_HorizontalKernel, 3, sizeof(cl_mem), (void*)&m_dKernelHorizontal);
	clError |= clSetKernelArg(m_HorizontalKernel, 4, sizeof(cl_uint), (void*)&m_Width);
	clError |= clSetKernelArg(m_HorizontalKernel, 5, sizeof(cl_uint), (void*)&m_Height);
	clError |= clSetKernelArg(m_HorizontalKernel, 6, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting horizontal kernel arguments");
		
	clError  = clSetKernelArg(m_VerticalKernel, 2, sizeof(cl_mem), (void*)&m_dDiscBuffer);
	clError |= clSetKernelArg(m_VerticalKernel, 3, sizeof(cl_mem), (void*)&m_dKernelVertical);
	clError |= clSetKernelArg(m_VerticalKernel, 4, sizeof(cl_uint), (void*)&m_Width);
	clError |= clSetKernelArg(m_VerticalKernel, 5, sizeof(cl_uint), (void*)&m_Height);
	clError |= clSetKernelArg(m_VerticalKernel, 6, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting vertical kernel arguments");

	return true;
}

void CConvolutionBilateral::ReleaseResources()
{

	// host data
	SAFE_DELETE_ARRAY( m_hCPUDiscBuffer );
	SAFE_DELETE_ARRAY( m_hGPUDiscBuffer );
	SAFE_DELETE_ARRAY( m_hNormDepthBuffer );
	for(int i = 0; i < 3; i++) {
		SAFE_DELETE_ARRAY( m_hSourceChannels[i] );
		SAFE_DELETE_ARRAY( m_hCPUResultChannels[i] );
		SAFE_DELETE_ARRAY( m_hGPUResultChannels[i] );

		SAFE_RELEASE_MEMOBJECT(m_dResultChannel[i]);
		SAFE_RELEASE_MEMOBJECT(m_dResultChannelTemp[i]);
	}

	SAFE_RELEASE_MEMOBJECT(m_dDiscBuffer);
	SAFE_RELEASE_MEMOBJECT(m_dNormDepthBuffer);

	SAFE_RELEASE_MEMOBJECT(m_dKernelHorizontal);
	SAFE_RELEASE_MEMOBJECT(m_dKernelVertical);

	SAFE_RELEASE_KERNEL(m_HorizontalDiscKernel);
	SAFE_RELEASE_KERNEL(m_VerticalDiscKernel);
	SAFE_RELEASE_KERNEL(m_HorizontalKernel);
	SAFE_RELEASE_KERNEL(m_VerticalKernel);
	SAFE_RELEASE_PROGRAM(m_Program);
}

void CConvolutionBilateral::ConvolutionChannelGPU(unsigned int iChannel, cl_context Context, cl_command_queue CommandQueue)
{
	cl_int clErr;

	clErr  = clSetKernelArg(m_HorizontalKernel, 0, sizeof(cl_mem), (void*)&m_dResultChannelTemp[iChannel]);
	clErr |= clSetKernelArg(m_HorizontalKernel, 1, sizeof(cl_mem), (void*)&m_dResultChannel[iChannel]);
	clErr |= clSetKernelArg(m_HorizontalKernel, 2, sizeof(cl_mem), (void*)&m_dDiscBuffer);
	V_RETURN_CL(clErr, "Error setting horizontal kernel arguments");

	size_t globalWorkSizeH[2] = {GetGlobalWorkSize(m_Width / m_StepsHorizontal, m_LocalSizeHorizontal[0]), GetGlobalWorkSize(m_Height, m_LocalSizeHorizontal[1])};	
	clErr = clEnqueueNDRangeKernel(CommandQueue, m_HorizontalKernel, 2, NULL, globalWorkSizeH, m_LocalSizeHorizontal, 0, NULL, NULL);
	V_RETURN_CL(clErr, "Error executing horizontal kernel!");

	clErr  = clSetKernelArg(m_VerticalKernel, 0, sizeof(cl_mem), (void*)&m_dResultChannel[iChannel]);
	clErr |= clSetKernelArg(m_VerticalKernel, 1, sizeof(cl_mem), (void*)&m_dResultChannelTemp[iChannel]);
	clErr |= clSetKernelArg(m_VerticalKernel, 2, sizeof(cl_mem), (void*)&m_dDiscBuffer);
	V_RETURN_CL(clErr, "Error setting vertical kernel arguments");

	size_t globalWorkSizeV[2] = {GetGlobalWorkSize(m_Width, m_LocalSizeVertical[0]), GetGlobalWorkSize(m_Height / m_StepsVertical, m_LocalSizeVertical[1])};
	clErr = clEnqueueNDRangeKernel(CommandQueue, m_VerticalKernel, 2, NULL, globalWorkSizeV, m_LocalSizeVertical, 0, NULL, NULL);
	V_RETURN_CL(clErr, "Error executing vertical kernel!");
}


//-------------------------------------------------------------------------------------------------------------------

void CConvolutionBilateral::ConvolutionChannelCPU(unsigned int Channel)
{
	//also measure the time for the first channel
	CTimer timer;


	//create a temporary data channel (this should be initialized once for all channels...)
	float* tempChannel = new float[m_Height * m_Pitch];

	timer.Start();

	// HORIZONTAL PASS
	for(int y = 0; y < m_Height; y++) 
	{
		for(int x = 0; x < m_Width; x++)
		{
			float sum = 0.f;
			float weight = 0.f;

			// Middle pixel
			weight	= m_hKernelHorizontal[m_KernelRadius];
			sum		= m_hSourceChannels[Channel][y * m_Pitch + x] * weight;

			// Left neighborhood
			for(int k = 0; k > -m_KernelRadius; ) {

				int flag = m_hCPUDiscBuffer[y * m_Pitch + x + k];
				// If discontinuity on the left detected, bail out
				if (flag & 1 ||  x+k <= 0)
					break;

				k--; 

				float w = m_hKernelHorizontal[m_KernelRadius - k];
				sum += m_hSourceChannels[Channel][y * m_Pitch + x + k] * w;
				weight += w;
			}

			// Right neighborhood
			for(int k = 0; k < m_KernelRadius; ) {

				int flag = m_hCPUDiscBuffer[y * m_Pitch + x + k];
				// If discontinuity on the right is detected, bail out
				if (flag & 2 || x+k >= m_Width-1)
					break;

				k++; 

				float w = m_hKernelHorizontal[m_KernelRadius - k];
				sum += m_hSourceChannels[Channel][y * m_Pitch + x + k] * w;
				weight += w;
			}

			// Re-normalize
			if (weight != 0.f)
				sum /= weight;
			else
				sum = 0.f;

			tempChannel[y * m_Pitch + x] = sum;
		}
	}

	//VERTICAL PASS
	for(int x = 0; x < m_Width; x++)
	{
		for(int y = 0; y < m_Height; y++)
		{
			float sum = 0.f;
			float weight = 0.f;

			// Middle pixel
			weight	= m_hKernelHorizontal[m_KernelRadius];
			sum		= tempChannel[y * m_Pitch + x] * weight;

			// Upper neighborhood
			for(int k = 0; k > -m_KernelRadius; ) {

				int flag = m_hCPUDiscBuffer[(y+k) * m_Pitch + x];
				// If discontinuity on the left detected, bail out
				if (flag & 4 || y+k <= 0)
					break;

				k--; 

				float w = m_hKernelHorizontal[m_KernelRadius - k];
				sum += tempChannel[(y+k) * m_Pitch + x] * w;
				weight += w;
			}

			// Lower neighborhood
			for(int k = 0; k < m_KernelRadius; ) {

				int flag = m_hCPUDiscBuffer[(y+k) * m_Pitch + x];
				// If discontinuity on the right is detected, bail out
				if (flag & 8 || y+k >= m_Height-1)
					break;

				k++; 

				float w = m_hKernelHorizontal[m_KernelRadius - k];
				sum += tempChannel[(y+k) * m_Pitch + x] * w;
				weight += w;
			}

			// Re-normalize
			if (weight != 0.f)
				sum /= weight;
			else
				sum = 0.f;

			m_hCPUResultChannels[Channel][y * m_Pitch + x] = sum;
		}
	}

	timer.Stop();

	if(Channel == 0)
	{
		double ms = 1000 * timer.GetElapsedTime();
		cout<<"  Average time: "<<ms<<" ms, throughput: "<< 1.0e-6 * (double)m_Width * m_Height / ms << " Gpixels/s" <<endl;
	}

	delete [] tempChannel;
}


void CConvolutionBilateral::ComputeCPU()
{

	// Detect discontinuities
	for(int y = 0; y < m_Height; y++)
		for(int x = 0; x < m_Width; x++)
		{
			cl_float4 myNormDepth = m_hNormDepthBuffer[y*m_Pitch + x];
			int flag = 0;

			// Left neighbor
			if (x > 0) {
				cl_float4 normDepth = m_hNormDepthBuffer[y*m_Pitch + x - 1];
				if (IsNormalDiscontinuity(myNormDepth, normDepth) || IsDepthDiscontinuity(myNormDepth.s[3], normDepth.s[3]))
					flag += 1;
			} else
				flag += 1;

			// Right neighbor
			if (x < m_Width - 1) {
				cl_float4 normDepth  = m_hNormDepthBuffer[y*m_Pitch + x + 1];
				if (IsNormalDiscontinuity(myNormDepth, normDepth) || IsDepthDiscontinuity(myNormDepth.s[3], normDepth.s[3]))
					flag += 2;
			} else
				flag += 2;

			// Upper neighbor
			if (y > 0) {
				cl_float4 normDepth = m_hNormDepthBuffer[(y-1)*m_Pitch + x];
				if (IsNormalDiscontinuity(myNormDepth, normDepth) || IsDepthDiscontinuity(myNormDepth.s[3], normDepth.s[3]))
					flag += 4;
			} else
				flag += 4;

			// Lower neighbor
			if (y < m_Height - 1) {
				cl_float4 normDepth  = m_hNormDepthBuffer[(y+1)*m_Pitch + x];
				if (IsNormalDiscontinuity(myNormDepth, normDepth) || IsDepthDiscontinuity(myNormDepth.s[3], normDepth.s[3]))
					flag += 8;
			} else
				flag += 8;

			m_hCPUDiscBuffer[y * m_Pitch + x] = flag;
		}

	// Do the convolution on the CPU
	for(unsigned int iChannel = 0; iChannel < 3; iChannel++)
		ConvolutionChannelCPU(iChannel);

	// Store CPU results
	SaveIntImage("CPUDiscontinuities.pfm", m_hCPUDiscBuffer, m_Width, m_Height, m_Pitch);
	SaveImage("CPUResultBilateral.pfm", m_hCPUResultChannels, m_Width, m_Height, m_Pitch);

}

//-------------------------------------------------------------------------------------------------------------------

void CConvolutionBilateral::ComputeGPU(cl_context Context, cl_command_queue CommandQueue)
{
	size_t dataSize = m_Width * m_Height * sizeof(cl_float);
	int nIterations = 10;

	cl_int clErr;

	// Copy input data to the GPU
	unsigned int numChannels = 3;
	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++) {
		V_RETURN_CL( clEnqueueWriteBuffer(CommandQueue, m_dResultChannel[iChannel], CL_FALSE, 0, dataSize, 
											m_hSourceChannels[iChannel], 0, NULL, NULL), "Error copying input data to device!" );
	}


	CTimer timer;
	clFinish(CommandQueue);
	timer.Start();

	// Detect discontinuities
	size_t globalWorkSizeH[2] = {GetGlobalWorkSize(m_Width / m_StepsHorizontal, m_LocalSizeHorizontal[0]), GetGlobalWorkSize(m_Height, m_LocalSizeHorizontal[1])};	
	clErr = clEnqueueNDRangeKernel(CommandQueue, m_HorizontalDiscKernel, 2, NULL, globalWorkSizeH, m_LocalSizeHorizontal, 0, NULL, NULL);
	V_RETURN_CL(clErr, "Error executing horizontal discontinuity kernel!");

	size_t globalWorkSizeV[2] = {GetGlobalWorkSize(m_Width, m_LocalSizeVertical[0]), GetGlobalWorkSize(m_Height / m_StepsVertical, m_LocalSizeVertical[1])};
	clErr = clEnqueueNDRangeKernel(CommandQueue, m_VerticalDiscKernel, 2, NULL, globalWorkSizeV, m_LocalSizeVertical, 0, NULL, NULL);
	V_RETURN_CL(clErr, "Error executing vertical discontinuity kernel!");


	// Perform the convolution
	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++) {
		ConvolutionChannelGPU(iChannel, Context, CommandQueue);
	}

	clFinish(CommandQueue);
	timer.Stop();

	double ms = 1000 * timer.GetElapsedTime() / double(nIterations);
	cout<<"  Average time: "<<ms<<" ms, throughput: "<< 1.0e-6 * (double)numChannels * m_Width * m_Height / ms << " Gpixels/s" <<endl;



	// Copy the results back to the CPU
	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++) {
		V_RETURN_CL( clEnqueueReadBuffer(CommandQueue, m_dResultChannel[iChannel], CL_TRUE, 0, dataSize,
			m_hGPUResultChannels[iChannel], 0, NULL, NULL), "Error reading back results from the device!" );
	}

	V_RETURN_CL( clEnqueueReadBuffer(CommandQueue, m_dDiscBuffer, CL_TRUE, 0, m_Width * m_Height * sizeof(int), 
		m_hGPUDiscBuffer, 0, NULL, NULL), "Error reading back results from the device!" );


	// Store the results
	SaveIntImage("GPUDiscontinuities.pfm", m_hGPUDiscBuffer, m_Width, m_Height, m_Pitch);
	SaveImage("GPUResultBilateral.pfm", m_hGPUResultChannels, m_Width, m_Height, m_Pitch);
}



bool CConvolutionBilateral::ValidateResults()
{
	//calculate the average squared difference
	float avgError = 0;
	float maxError = 0;
	float scaling = 1.0f / float(3 * m_Width * m_Height);

	for(int y = 0; y < m_Height; y++)
	{
		for(int x = 0; x < m_Width; x++)
			for(int i = 0; i < 3; i++)
			{
				float L2Error = m_hCPUResultChannels[i][y * m_Pitch + x] - m_hGPUResultChannels[i][y * m_Pitch + x];
				L2Error = L2Error * L2Error;
				maxError = max(maxError, L2Error);
				avgError += L2Error * scaling;

				//temp: to see the difference
				m_hCPUResultChannels[i][y * m_Pitch + x] = L2Error;
			}
	}
	cout<<"Average sq. difference: "<<avgError<<endl;
	cout<<"Maximum sq. difference: "<<maxError<<endl;

	//save difference image
	SaveImage("DifferenceImageBilateral.pfm", m_hCPUResultChannels, m_Width, m_Height, m_Pitch);

	return avgError < 0.0001f;
}
