#include "CConvolutionSeparable.h"
#include "Common.h"
#include "CTimer.h"

#include "Pfm.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CConvolutionSeparable

CConvolutionSeparable::CConvolutionSeparable(const char* FileName, size_t LocalSizeHorizontal[2], size_t LocalSizeVertical[2],
		int StepsHorizontal, int StepsVertical, int KernelRadius, float* pKernelHorizontal, float* pKernelVertical)
		:m_FileName(FileName),
		m_Height(0), m_Width(0), m_KernelRadius(KernelRadius), m_StepsHorizontal(StepsHorizontal), m_StepsVertical(StepsVertical),
		m_dKernelHorizontal(NULL), m_dKernelVertical(NULL),
		m_Program(NULL), m_HorizontalKernel(NULL), m_VerticalKernel(NULL)
{
	m_LocalSizeHorizontal[0] = LocalSizeHorizontal[0];
	m_LocalSizeHorizontal[1] = LocalSizeHorizontal[1];
	m_LocalSizeVertical[0] = LocalSizeVertical[0];
	m_LocalSizeVertical[1] = LocalSizeVertical[1];

	const unsigned int kernelSize = 2 * m_KernelRadius + 1;
	m_hKernelHorizontal = new float[kernelSize];
	m_hKernelVertical = new float[kernelSize];
	memcpy(m_hKernelHorizontal, pKernelHorizontal, kernelSize * sizeof(float));
	memcpy(m_hKernelVertical, pKernelVertical, kernelSize * sizeof(float));

	for(int i = 0; i < 3; i++)
	{
		m_hSourceChannels[i] = NULL;
		m_hCPUResultChannels[i] = NULL;
		m_hGPUResultChannels[i] = NULL;
		m_dGPUInputBuffers[i] = NULL;
		m_dGPUOutputBuffers[i] = NULL;
	}

	m_dGPUWorkingBuffer = NULL;
}

CConvolutionSeparable::~CConvolutionSeparable()
{
	delete [] m_hKernelHorizontal;
	delete [] m_hKernelVertical;
}

bool CConvolutionSeparable::InitResources(cl_device_id Device, cl_context Context)
{
	PFM inputPfm;
	if (!inputPfm.LoadRGB(m_FileName.c_str())) {
		cout<<"Error loading file: " << m_FileName.c_str() << "." << endl;
		return false;
	}

	//internally, we convert the bitmap to floats, and execute the same convolution
	//operation on its three channels separately
	m_Height = inputPfm.height;
	m_Width = inputPfm.width;
	m_Pitch = m_Width;
	if(m_Width % 32 != 0)
		m_Pitch = m_Width + 32 - (m_Width % 32); //This will make sure that the data accesses are ALWAYS coalesced

	cout<<"Size of image: "<<m_Width<<" x "<<m_Height<<endl;

	//allocate data for the float channels
	for(int i = 0; i < 3; i++)
	{
		m_hSourceChannels[i] = new float[m_Height * m_Pitch];
		m_hCPUResultChannels[i] = new float[m_Height * m_Pitch];
		m_hGPUResultChannels[i] = new float[m_Height * m_Pitch];
	}

	unsigned int pixelOffset = 0;
	unsigned int trippleOffset = 0;
	for(int y = 0; y < m_Height; y++) {
		for(int x = 0; x < m_Width; x++) {
			m_hSourceChannels[0][pixelOffset] = inputPfm.pImg[trippleOffset    ];
			m_hSourceChannels[1][pixelOffset] = inputPfm.pImg[trippleOffset + 1];
			m_hSourceChannels[2][pixelOffset] = inputPfm.pImg[trippleOffset + 2];

			pixelOffset++;
			trippleOffset += 3;
		}
		pixelOffset += m_Pitch - m_Width;
	}

	//create GPU resources
	//we can init the kernel buffer during creation as its contents will not change
	const unsigned int kernelSize = 2 * m_KernelRadius + 1;

	cl_int clError = 0;
	cl_int clErr;
	m_dKernelHorizontal = clCreateBuffer(Context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, kernelSize * sizeof(cl_float), 
		m_hKernelHorizontal, &clErr);
	clError |= clErr;
	m_dKernelVertical = clCreateBuffer(Context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, kernelSize * sizeof(cl_float), 
		m_hKernelVertical, &clErr);
	clError |= clErr;
	V_RETURN_FALSE_CL(clError, "Error allocating device kernel constants.");

	for(int i = 0; i < 3; i++)
	{
		m_dGPUInputBuffers[i] = clCreateBuffer(Context, CL_MEM_READ_ONLY, m_Pitch * m_Height * sizeof(cl_float), NULL, &clError);
		V_RETURN_FALSE_CL(clError, "Error allocating device input array");
		m_dGPUOutputBuffers[i] = clCreateBuffer(Context, CL_MEM_WRITE_ONLY, m_Pitch * m_Height * sizeof(cl_float), NULL, &clError);
		V_RETURN_FALSE_CL(clError, "Error allocating device output array");
	}
	m_dGPUWorkingBuffer = clCreateBuffer(Context, CL_MEM_READ_WRITE, m_Pitch * m_Height * sizeof(cl_float), NULL, &clError);
	V_RETURN_FALSE_CL(clError, "Error allocating device working array");

	char* programCode = NULL;
	size_t programSize = 0;

	LoadProgram("ConvolutionSeparable.cl", &programCode, &programSize);

	//create a program object
	m_Program = clCreateProgramWithSource(Context, 1, (const char**) &programCode, &programSize, &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create program from file.");

	//build program
	
	//This time we define several kernel-specific constants that we did not know during
	//implementing the kernel, but we need to include during compile time.
	char compileOptions[2048];
	sprintf_s(compileOptions, "-cl-fast-relaxed-math " \
		"-D KERNEL_RADIUS=%d " \
		"-D H_GROUPSIZE_X=%d -D H_GROUPSIZE_Y=%d -D H_RESULT_STEPS=%d " \
		"-D V_GROUPSIZE_X=%d -D V_GROUPSIZE_Y=%d -D V_RESULT_STEPS=%d ",
		m_KernelRadius, 
		m_LocalSizeHorizontal[0], m_LocalSizeHorizontal[1], m_StepsHorizontal,
		m_LocalSizeVertical[0], m_LocalSizeVertical[1], m_StepsVertical);
	
	clError = clBuildProgram(m_Program, 1, &Device, compileOptions, NULL, NULL);
	if(clError != CL_SUCCESS)
	{
		PrintBuildLog(m_Program, Device);
		return false;
	}

	//create kernel(s)
	m_HorizontalKernel = clCreateKernel(m_Program, "ConvHorizontal", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create horizontal kernel.");
	
	m_VerticalKernel = clCreateKernel(m_Program, "ConvVertical", &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create vertical kernel.");

	//bind kernel attributes
	//the resulting image will be in buffer 1
	clError = clSetKernelArg(m_HorizontalKernel, 2, sizeof(cl_mem), (void*)&m_dKernelHorizontal);
	clError |= clSetKernelArg(m_HorizontalKernel, 3, sizeof(cl_uint), (void*)&m_Width);
	clError |= clSetKernelArg(m_HorizontalKernel, 4, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting horizontal kernel arguments");
		
	//the resulting image will be in buffer 0
	clError = clSetKernelArg(m_VerticalKernel, 2, sizeof(cl_mem), (void*)&m_dKernelVertical);
	clError |= clSetKernelArg(m_VerticalKernel, 3, sizeof(cl_uint), (void*)&m_Height);
	clError |= clSetKernelArg(m_VerticalKernel, 4, sizeof(cl_uint), (void*)&m_Pitch);
	V_RETURN_FALSE_CL(clError, "Error setting vertical kernel arguments");

	return true;
}

void CConvolutionSeparable::ReleaseResources()
{
	// host data
	for(int i = 0; i < 3; i++)
	{
		SAFE_DELETE_ARRAY( m_hSourceChannels[i] );
		SAFE_DELETE_ARRAY( m_hCPUResultChannels[i] );
		SAFE_DELETE_ARRAY( m_hGPUResultChannels[i] );
		SAFE_RELEASE_MEMOBJECT(m_dGPUInputBuffers[i]);
		SAFE_RELEASE_MEMOBJECT(m_dGPUOutputBuffers[i]);
	}

	SAFE_RELEASE_MEMOBJECT(m_dGPUWorkingBuffer);
	SAFE_RELEASE_MEMOBJECT(m_dKernelHorizontal);
	SAFE_RELEASE_MEMOBJECT(m_dKernelVertical);

	SAFE_RELEASE_KERNEL(m_HorizontalKernel);
	SAFE_RELEASE_KERNEL(m_VerticalKernel);
	SAFE_RELEASE_PROGRAM(m_Program);
}

void CConvolutionSeparable::ConvolutionChannelGPU(unsigned int Channel, cl_context Context, 
	cl_command_queue CommandQueue, int NIterations)
{
	cl_int clErr;

	clErr  = clSetKernelArg(m_HorizontalKernel, 0, sizeof(cl_mem), (void*)&m_dGPUWorkingBuffer);
	clErr |= clSetKernelArg(m_HorizontalKernel, 1, sizeof(cl_mem), (void*)&m_dGPUInputBuffers[Channel]);
	V_RETURN_CL(clErr, "Error setting horizontal kernel arguments");

	clErr  = clSetKernelArg(m_VerticalKernel, 0, sizeof(cl_mem), (void*)&m_dGPUOutputBuffers[Channel]);
	clErr |= clSetKernelArg(m_VerticalKernel, 1, sizeof(cl_mem), (void*)&m_dGPUWorkingBuffer);
	V_RETURN_CL(clErr, "Error setting vertical kernel arguments");

	for(int i = 0; i < NIterations; i++)
	{

		size_t globalWorkSizeH[2] = {GetGlobalWorkSize(m_Width / m_StepsHorizontal, m_LocalSizeHorizontal[0]), GetGlobalWorkSize(m_Height, m_LocalSizeHorizontal[1])};	
		clErr = clEnqueueNDRangeKernel(CommandQueue, m_HorizontalKernel, 2, NULL, globalWorkSizeH, m_LocalSizeHorizontal, 0, NULL, NULL);
		V_RETURN_CL(clErr, "Error executing horizontal kernel!");

		size_t globalWorkSizeV[2] = {GetGlobalWorkSize(m_Width, m_LocalSizeVertical[0]), GetGlobalWorkSize(m_Height / m_StepsVertical, m_LocalSizeVertical[1])};
		clErr = clEnqueueNDRangeKernel(CommandQueue, m_VerticalKernel, 2, NULL, globalWorkSizeV, m_LocalSizeVertical, 0, NULL, NULL);
		V_RETURN_CL(clErr, "Error executing vertical kernel!");
	
	}

}

void CConvolutionSeparable::ComputeGPU(cl_context Context, cl_command_queue CommandQueue)
{
	size_t dataSize = m_Pitch * m_Height * sizeof(cl_float);
	int nIterations = 10;

	unsigned int numChannels = 3;
	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)
	{
		//copy input data to the GPU
		V_RETURN_CL( clEnqueueWriteBuffer(CommandQueue, m_dGPUInputBuffers[iChannel], CL_FALSE, 0, dataSize, 
												m_hSourceChannels[iChannel], 0, NULL, NULL), "Error copying input data to device!" );

	}

	CTimer timer;
	
	clFinish(CommandQueue);
	timer.Start();

	//perform the convolution
	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)
	{
		ConvolutionChannelGPU(iChannel, Context, CommandQueue, nIterations);
	}

	clFinish(CommandQueue);
	timer.Stop();

	double ms = 1000 * timer.GetElapsedTime() / double(nIterations);
	cout<<"  Average time: "<<ms<<" ms, throughput: "<< 1.0e-6 * (double)numChannels * m_Width * m_Height / ms << " Gpixels/s" <<endl;

	for(unsigned int iChannel = 0; iChannel < numChannels; iChannel++)
	{
		//copy the results back to the CPU
		//(this time the data is in the same buffer as the input was, because of the 2 convolution passes)
		V_RETURN_CL( clEnqueueReadBuffer(CommandQueue, m_dGPUOutputBuffers[iChannel], CL_TRUE, 0, dataSize,
									m_hGPUResultChannels[iChannel], 0, NULL, NULL), "Error reading back results from the device!" );

	}
	
	SaveImage("GPUResultSeparable.pfm", m_hGPUResultChannels, m_Width, m_Height, m_Pitch);
}

void CConvolutionSeparable::ConvolutionChannelCPU(unsigned int Channel)
{
	//also measure the time for the first channel
	CTimer timer;
	
	
	//create a temporary data channel
	float* tempChannel = new float[m_Height * m_Pitch];
	
	timer.Start();

	//horizontal pass
	for(int y = 0; y < m_Height; y++)
		for(int x = 0; x < m_Width; x++)
		{
			float value = 0;
			//apply horizontal kernel
			for(int k = -m_KernelRadius; k <= m_KernelRadius; k++)
			{
				int sx = x + k;
				if(sx >= 0 && sx < m_Width)
					value += m_hSourceChannels[Channel][y * m_Pitch + sx] * m_hKernelHorizontal[m_KernelRadius - k];
			}
			tempChannel[y * m_Pitch + x] = value;
		}

	//vertical pass
	for(int x = 0; x < m_Width; x++)
		for(int y = 0; y < m_Height; y++)
		{
			float value = 0;
			//apply horizontal kernel
			for(int k = -m_KernelRadius; k <= m_KernelRadius; k++)
			{
				int sy = y + k;
				if(sy >= 0 && sy < m_Height)
					value += tempChannel[sy * m_Pitch + x] * m_hKernelVertical[m_KernelRadius - k];
			}
			m_hCPUResultChannels[Channel][y * m_Pitch + x] = value;
		}

	timer.Stop();

	if(Channel == 0)
	{
		double ms = 1000 * timer.GetElapsedTime();
		cout<<"  Average time: "<<ms<<" ms, throughput: "<< 1.0e-6 * (double)m_Width * m_Height / ms << " Gpixels/s" <<endl;
	}

	delete [] tempChannel;

}

void CConvolutionSeparable::ComputeCPU()
{
	//do the convolution on the CPU

	for(unsigned int iChannel = 0; iChannel < 3; iChannel++)
		ConvolutionChannelCPU(iChannel);

	SaveImage("CPUResultSeparable.pfm", m_hCPUResultChannels, m_Width, m_Height, m_Pitch);
}

bool CConvolutionSeparable::ValidateResults()
{
	//calculate the average squared difference
	float avgError = 0;
	float maxError = 0;
	float scaling = 1.0f / float(3 * m_Width * m_Height);

	for(int y = 0; y < m_Height; y++)
	{
		for(int x = 0; x < m_Width; x++)
			for(int i = 0; i < 3; i++)
			{
				float L2Error = m_hCPUResultChannels[i][y * m_Pitch + x] - m_hGPUResultChannels[i][y * m_Pitch + x];
				L2Error = L2Error * L2Error;
				maxError = max(maxError, L2Error);
				avgError += L2Error * scaling;

				//temp: to see the difference
				m_hCPUResultChannels[i][y * m_Pitch + x] = L2Error;
			}
	}
	cout<<"Average sq. difference: "<<avgError<<endl;
	cout<<"Maximum sq. difference: "<<maxError<<endl;

	//save difference image
	SaveImage("DifferenceImageSeparable.pfm", m_hCPUResultChannels, m_Width, m_Height, m_Pitch);

	return avgError < 0.0001f;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
