// All OpenCL headers
#if defined (__APPLE__) || defined(MACOSX)
    #include <OpenCL/opencl.h>
#else
    #include <CL/opencl.h>
#endif 

#include "Common.h"
#include "IAssignment.h"

#include "CConvolution3x3.h"
#include "CConvolutionSeparable.h"
#include "CConvolutionBilateral.h"

using namespace std;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Global variables and forward declarations

cl_context			g_CLContext			= NULL;
cl_command_queue	g_CLCommandQueue	= NULL;
cl_platform_id		g_CLPlatform		= NULL;
cl_device_id		g_CLDevice			= NULL;

bool InitContextResources();
void CleanupContextResources();
bool RunAssigment(IAssignment& Assignment);

//Uncomment one of these to test the separable convolution for various kernels
// (of course feel free to define your own kernels)

//Gaussian blur with radius 3 (7x7)
//#define GAUSSIAN_BLUR3
//Box filter with radius 4 (9x9)
//#define BOX_BLUR4
//Box filter with radius 8 (17x17)
#define BOX_BLUR8

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv)
{
	if(InitContextResources())
	{
		cout<<"OpenCL context initialized."<<endl;

		cout<<"########################################"<<endl;
		cout<<"ASSIGNMENT 1: 3x3 convolution"<<endl<<endl;
		{
			size_t TileSize[2] = {32, 16};
			float ConvKernel[3][3] = {
				{-1, -1, 0},
				{-1, 0, 1},
				{0, 1, 1},
			};
			RunAssigment(CConvolution3x3("input.pfm", TileSize,
				ConvKernel, true, 0.5f));
		}
		
		cout<<endl<<"########################################"<<endl;
		cout<<"ASSIGNMENT 2: Separable convolution"<<endl<<endl;
		{
			size_t HGroupSize[2] = {32, 16};
			size_t VGroupSize[2] = {32, 16};

		
			#ifdef BOX_BLUR4
				//simple box filter
				float ConvKernel[9];
				for(int i = 0; i < 9; i++)
					ConvKernel[i] = 1.0f / 9.0f;

				RunAssigment(CConvolutionSeparable("input.pfm", HGroupSize, VGroupSize,
					4, 4, 4, ConvKernel, ConvKernel));
			#endif

			#ifdef BOX_BLUR8
				//simple box filter
				float ConvKernel[17];
				for(int i = 0; i < 17; i++)
					ConvKernel[i] = 1.0f / 17.0f;

				RunAssigment(CConvolutionSeparable("input.pfm", HGroupSize, VGroupSize,
					4, 4, 8, ConvKernel, ConvKernel));
			#endif

			#ifdef GAUSSIAN_BLUR3
				// Gaussian blur
				float ConvKernel[7] = {0.000817774f, 0.0286433f, 0.235018f, 0.471041f,
										0.235018f, 0.0286433f, 0.000817774f};
				RunAssigment(CConvolutionSeparable("input.pfm", HGroupSize, VGroupSize,
					4, 4, 3, ConvKernel, ConvKernel));
			#endif
		}

		cout<<endl<<"########################################"<<endl;
		cout<<"ASSIGNMENT 3: Separable bilateral convolution"<<endl<<endl;
		{
			size_t HGroupSize[2] = {32, 4};
			size_t VGroupSize[2] = {32, 4};

			float ConvKernel[9] = {0.010284844f,	0.0417071f,	0.113371652f,	0.206576619f,	0.252313252f,	0.206576619f,	0.113371652f,	0.0417071f,	0.010284844f};

			RunAssigment(CConvolutionBilateral("color.pfm", "normals.pfm", "depth.pfm", HGroupSize, VGroupSize, 4, 4, 4, ConvKernel, ConvKernel));
		}
	}

	CleanupContextResources();
	
	cout<<"Press any key..."<<endl;
	cin.get();
}

bool InitContextResources()
{
	//error code
	cl_int clError;
	
	//get platform ID
	V_RETURN_FALSE_CL( clGetPlatformIDs(1, &g_CLPlatform, NULL), "Failed to get CL platform ID" );

	//get a reference to the first available GPU device
	V_RETURN_FALSE_CL( clGetDeviceIDs(g_CLPlatform, CL_DEVICE_TYPE_GPU, 1, &g_CLDevice, NULL), "No GPU device found." );

	char deviceName[256];
	V_RETURN_FALSE_CL( clGetDeviceInfo(g_CLDevice, CL_DEVICE_NAME, 256, &deviceName, NULL), "Unable to query device name.");
	cout << "Device: " << deviceName << endl;

	//Create a new OpenCL context on the selected device
	g_CLContext = clCreateContext(0, 1, &g_CLDevice, NULL, NULL, &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create OpenCL context.");

	//Finally, create the command queue. All the asynchronous commands to the device will be issued
	//from the CPU into this queue. This way the host program can continue the execution until some results
	//from that device are needed.
	g_CLCommandQueue = clCreateCommandQueue(g_CLContext, g_CLDevice, 0, &clError);
	V_RETURN_FALSE_CL(clError, "Failed to create the command queue in the context");

	return true;
}

void CleanupContextResources()
{
	if(g_CLCommandQueue)	clReleaseCommandQueue(g_CLCommandQueue);
	if(g_CLContext)			clReleaseContext(g_CLContext);
}

bool RunAssigment(IAssignment& pAssignment)
{
	if(!pAssignment.InitResources(g_CLDevice, g_CLContext))
	{
		cout<<"Error during resource allocation. Aborting execution."<<endl;
		pAssignment.ReleaseResources();
		return false;
	}

	//compute golden result
	cout<<"Computing CPU reference result..."<<endl;
	pAssignment.ComputeCPU();

	//running the same task on the GPU
	cout<<endl<<"Computing GPU result..."<<endl;
	
	//run  the kernel.
	pAssignment.ComputeGPU(g_CLContext, g_CLCommandQueue);

	//validate results
	if(pAssignment.ValidateResults())
		cout<<"CPU REFERENCE TEST PASSED!"<<endl;
	else
		cout<<"INVALID RESULTS!"<<endl;

	//cleanup
	pAssignment.ReleaseResources();

	return true;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////


